"""This module communicates with SigmaStudio.

It can receive read/write requests and return with read response packets.
"""
import logging
import socket
import socketserver
import threading
from multiprocessing import Pipe
from typing import Type

from sigmadsp.communication.base import (
    ReadRequest,
    SafeloadRequest,
    ThreadedTCPServer,
    WriteRequest,
)
from sigmadsp.communication.sigmastudio_protocols import (
    SigmaProtocolHeader,
    SigmaProtocolPacket,
)

# A logger for this module
logger = logging.getLogger(__name__)


class SigmaStudioRequestHandler(socketserver.BaseRequestHandler):
    """Request handler for messages from SigmaStudio."""

    request: socket.socket
    server: ThreadedTCPServer
    dsp_type: str

    def read(self, amount: int) -> bytearray:
        """Reads the specified amount of data from the socket.

        Args:
            amount (int): The number of bytes to get.

        Returns:
            bytearray: The received data.
        """
        data = bytearray()

        while amount > len(data):
            # Wait until the complete TCP payload was received.
            received = self.request.recv(amount - len(data))

            if 0 == len(received):
                # Give up, if no more data arrives.
                # Close the socket.
                self.request.shutdown(socket.SHUT_RDWR)
                self.request.close()

                raise ConnectionError

            data.extend(received)

        return data

    def handle_write_data(self, packet: SigmaProtocolPacket):
        """Handle requests, where SigmaStudio wants to write to the DSP.

        Args:
            packet (SigmaProtocolPacket): The request packet object.
        """
        request: WriteRequest

        if packet.header.is_safeload:
            logger.info("[safeload] %s bytes to address 0x%04x", packet.header["data_length"], packet.header["address"])
            request = SafeloadRequest(packet.header["address"], packet.payload)
        else:
            logger.info("[write] %s bytes to address 0x%04x", packet.header["data_length"], packet.header["address"])
            request = WriteRequest(packet.header["address"], packet.payload)

        self.server.pipe_end_owner.send(request)

    def handle_read_request(self, packet: SigmaProtocolPacket):
        """Handle requests, where SigmaStudio wants to read from the DSP.

        Args:
            packet (SigmaProtocolPacket): The request header object.
        """
        logger.info("[read] %s bytes from address 0x%04x", packet.header["data_length"], packet.header["address"])

        # Notify application of read request
        self.server.pipe_end_owner.send(ReadRequest(packet.header["address"], packet.header["data_length"]))

        # Wait for payload data that goes into the read response
        read_response = self.server.pipe_end_owner.recv()

        header_defaults = packet.header.fields

        response_packet = SigmaProtocolPacket(self.dsp_type)
        response_packet.init_from_payload(SigmaProtocolHeader.READ_RESPONSE, read_response.data, header_defaults)
        response_packet.header["success"] = 0

        self.request.sendall(response_packet.as_bytes)

    def handle(self):
        """Call, when the TCP server receives new data for handling.

        It never stops, except if the connection is reset.
        """
        while True:
            try:
                packet: SigmaProtocolPacket = SigmaProtocolPacket(self.dsp_type)
                packet.init_from_network(self)

                if packet.header.is_write_request:
                    self.handle_write_data(packet)
                elif packet.header.is_read_request:
                    self.handle_read_request(packet)

            except ConnectionError:
                break


class Adau14xxRequestHandler(SigmaStudioRequestHandler):
    """ADAU144x/5x/6x request handler."""

    dsp_type = "adau14xx"


class Adau1701RequestHandler(SigmaStudioRequestHandler):
    """ADAU1701 request handler."""

    dsp_type = "adau1701"


class SigmaStudioInterface:
    """This is an interface class for communicating with SigmaStudio.

    It creates a TCP server, which SigmaStudio talks to.
    """

    def __init__(self, host: str, port: int, dsp_type: str):
        """Initialize the SigmaStudio interface.

        Starts the main TCP worker and initializes a pipe for communicating with the sigmadsp backend.

        Args:
            host (str): Listening IP address
            port (int): Port to listen at
            dsp_type (str): DSP type, used to select the relevant protocol handler
        """
        self.host = host
        self.port = port
        self.dsp_type = dsp_type

        # Generate a Pipe for communicating with the TCP server worker thread within this class.
        self.pipe_end_owner, self.pipe_end_user = Pipe()

        tcp_server_worker_thread = threading.Thread(target=self.tcp_server_worker, name="TCP server worker thread")
        tcp_server_worker_thread.daemon = True
        tcp_server_worker_thread.start()

    def tcp_server_worker(self):
        """The main worker for the TCP server."""
        protocol_handler: Type[SigmaStudioRequestHandler]

        if self.dsp_type == "adau14xx":
            protocol_handler = Adau14xxRequestHandler

        elif self.dsp_type == "adau1701":
            protocol_handler = Adau1701RequestHandler

        else:
            raise TypeError(f"The specified DSP type {self.dsp_type} is not supported.")

        tcp_server = ThreadedTCPServer((self.host, self.port), protocol_handler)

        with tcp_server:
            # Base TCP server thread
            # This initial thread starts one more thread for each request.
            tcp_server_thread = threading.Thread(target=tcp_server.serve_forever, name="TCP server thread")
            tcp_server_thread.daemon = True
            tcp_server_thread.start()

            while True:
                # Wait for a request from the TCP server
                request = tcp_server.pipe_end_user.recv()

                if isinstance(request, WriteRequest):
                    # Write request received, don't do anything else
                    self.pipe_end_owner.send(request)

                elif isinstance(request, SafeloadRequest):
                    # Safeload request received, don't do anything else
                    self.pipe_end_owner.send(request)

                elif isinstance(request, ReadRequest):
                    # Read request received, wait for data to send to PC application
                    self.pipe_end_owner.send(request)

                    read_response = self.pipe_end_owner.recv()

                    # Send read response
                    tcp_server.pipe_end_user.send(read_response)



"""This module provides functionality for controlling SigmaDSP hardware, e.g.

- Deploying programs
- Changing parameter register contents
- Reading parameter registers
"""
import logging
from typing import Union

from sigmadsp.hardware.dsp import Dsp
from sigmadsp.helper.conversion import (
    bytes_to_int16,
    bytes_to_int32,
    float_to_frac_5_23,
    frac_5_23_to_float,
    int16_to_bytes,
    int32_to_bytes,
)

# A logger for this module
logger = logging.getLogger(__name__)


class Adau1701(Dsp):
    """A class for controlling functionality of Analog Devices Sigma DSPs, especially ADAU1701 series parts."""

    # Addresses and sizes of important registers
    CONTROL_REGISTER = 0x081C
    CONTROL_REGISTER_LENGTH = 2

    # All fixpoint (parameter) registers are four bytes long
    FIXPOINT_REGISTER_LENGTH = 4

    # Safeload registers (address, data)
    SAFELOAD_REGISTERS = [
        (0x0815, 0x810),
        (0x0816, 0x811),
        (0x0815, 0x812),
        (0x0818, 0x813),
        (0x0819, 0x814),
    ]

    # Safeload register length
    SAFELOAD_SA_LENGTH = 2
    SAFELOAD_SD_LENGTH = 5

    def soft_reset(self):
        """Soft reset the DSP.

        Not available on ADAU1701
        """
        logger.info("Soft-resetting the DSP is not available on ADAU1701")

    def get_parameter_value(self, address: int, data_format: str) -> Union[float, int, None]:
        """Get a parameter value from a chosen register address.

        Args:
            address (int): The address to look at.
            data_format (str): The data type to return the register in. Can be 'float' or 'int'.

        Returns:
            Union[float, int, None]: Representation of the register content in the specified format.
        """
        data_register = self.read(address, Adau1701.FIXPOINT_REGISTER_LENGTH)
        data_integer = bytes_to_int32(data_register)

        float_value = frac_5_23_to_float(data_integer)

        if "int" == data_format:
            return int(float_value)

        elif "float" == data_format:
            return float_value

        else:
            return None

    def set_parameter_value(self, value: Union[float, int], address: int) -> None:
        """Set a parameter value for a chosen register address.

        Args:
            value (float): The value to store in the register
            address (int): The target address
        """
        data_register: Union[bytes, None] = None
        data_register = int32_to_bytes(float_to_frac_5_23(value))

        if data_register is not None:
            self.safeload(address, data_register)

    def safeload(self, address: int, data: bytes, count: int = 1):
        """Write data to the chip using hardware safeload.

        Args:
            address (int): Address to write to
            data (bytes): Data to write; multiple words should be concatenated
            count (int): number of words to write (max. 5)
        """
        # load up the address and data in safeload registers
        for sd in range(0, count):
            address_register, data_register = Adau1701.SAFELOAD_REGISTERS[sd]
            address_bytes = int16_to_bytes(address)
            data_buf = bytearray(Adau1701.SAFELOAD_SD_LENGTH)

            data_buf[1:] = data[sd * Adau1701.FIXPOINT_REGISTER_LENGTH : (sd + 1) * Adau1701.FIXPOINT_REGISTER_LENGTH]

            self.write(address_register, address_bytes)
            self.write(data_register, data_buf)

        control_bytes = self.read(Adau1701.CONTROL_REGISTER, Adau1701.CONTROL_REGISTER_LENGTH)
        control_reg = bytes_to_int16(control_bytes)

        ist_mask = 1 << 5

        control_reg |= ist_mask

        # start safe load
        self.write(Adau1701.CONTROL_REGISTER, int16_to_bytes(control_reg))



"""This module contains implementations for SigmaStudio TCPIP* protocol decoding.

These handle the packet decoding and conditioning, while actual communication happens upstream.

Protocol headers for ADAU144x/5x/6x are documented on the Analog Devices wiki at
https://wiki.analog.com/resources/tools-software/sigmastudio/usingsigmastudio/tcpipchannels

Protocol headers for ADAU1401/1701 are documented on the Analog Devices forum at
https://ez.analog.com/dsp/sigmadsp/f/q-a/163849/adau1401-adau1701-tcp-ip-documentation-unonficial-self-made
"""
from abc import ABC
from collections import OrderedDict
from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, Iterator, List, Type, Union

from sigmadsp.helper.conversion import bytes_to_int, int_to_bytes

if TYPE_CHECKING:
    # avoid circular import
    from sigmadsp.communication.sigma_tcp_server import SigmaStudioRequestHandler


@dataclass
class Field:
    """A class that represents a field in a protocol header."""

    # The name of the field, e.g. "operation", "safeload", "channel".
    name: str

    # The offset of the field in bytes from the start of the header.
    offset: int

    # The size of the field in bytes.
    size: int

    # The stored value.
    value: int = 0

    def __post_init__(self):
        """Perform sanity checks on the field properties."""
        if self.size < 0:
            raise ValueError("Field size must be a positive integer.")

        if self.offset < 0:
            raise ValueError("Field offset must be a positive integer.")

        # The last byte index that is occupied by this field.
        self.end = self.offset + self.size - 1

    def __hash__(self) -> int:
        """Hash functionality."""
        return hash((self.name, self.offset, self.size))


class Fields:
    """An iterable collection of Field objects."""

    def __init__(self, fields: List[Field]):
        """Initialize the header fields. Add more fields to it by means of `add()`.

        Instantiate via:
        Fields(
            [
                Field("field_name", 0, 4),
                Field("next_field_name", 4, 1),
                # ...
            ]
        )

        Args:
            fields (List[Field]): The list of fields to add initially.
        """
        self._fields: OrderedDict[str, Field] = OrderedDict()  # pylint: disable=E1136

        for field in fields:
            self.add_field(field)

    @property
    def size(self) -> int:
        """The total size of the header in bytes."""
        return sum([field.size for field in self])

    @property
    def is_continuous(self) -> bool:
        """Whether or not there are spaces in the header that are not defined."""
        fields_entries = self.as_list()

        for field, next_field in zip(fields_entries, fields_entries[1:]):
            if (field.end + 1) != next_field.offset:
                return False

        return True

    def _check_for_overlaps(self):
        """Check for overlapping fields.

        Raises:
            MemoryError: If overlapping fields are found.
        """
        fields_entries = self.as_list()

        # Check for overlapping fields, which are sorted by their offset.
        for field, next_field in zip(fields_entries, fields_entries[1:]):
            if not field.end <= next_field.offset:
                raise MemoryError("Fields {field.name} and {next_field.name} overlap.")

    def _sort_fields_by_offset(self):
        """Sorts the fields in this header by their offset."""
        self._fields = OrderedDict(sorted(self._fields.items(), key=lambda item: item[1].offset))

    def add_field(self, field: Field):
        """Add a new field. Duplicates are ignored.

        Args:
            field (Field): The field to add.
        """
        if field not in self:
            self._fields[field.name] = field

            self._sort_fields_by_offset()
            self._check_for_overlaps()

    def as_bytes(self) -> bytes:
        """Get the full header as a bytes object."""
        buffer = bytearray()

        for field in self:
            int_to_bytes(field.value, buffer, field.offset, field.size)

        return bytes(buffer)

    def as_list(self) -> List[Field]:
        """The fields as a list.

        Returns:
            List[Field]: The list of fields.
        """
        return list(self._fields.values())

    def __iter__(self) -> Iterator[Field]:
        """The iterator for fields."""
        for item in self._fields.values():
            yield item

    def __getitem__(self, name: str) -> Field:
        """Get a field by its name.

        Args:
            name (str): The name of the field.

        Returns:
            Union[Field, None]: The field, or None, if no field was found.

        Raises:
            IndexError: If the field does not exist.
        """
        return self._fields[name]

    def __contains__(self, name: str) -> bool:
        """Magic methods for using `in`.

        Args:
            name (str): The name to look for in the fields.

        Returns:
            bool: True, if a field with the given name exists.
        """
        return name in self._fields


class SigmaProtocolHeader(ABC):
    """Base class for protocol headers."""

    READ_REQUEST: int = 0x0A
    READ_RESPONSE: int = 0x0B
    WRITE: int = 0x09

    @property
    def fields(self) -> Fields:
        """Fields of the header."""

    @property
    def has_payload(self) -> bool:
        """Whether this kind of packet carries a payload.

        Returns:
            bool: True if this is a write / read response, False otherwise.
        """
        return self.fields["operation"].value in [self.WRITE, self.READ_RESPONSE]

    @property
    def is_write_request(self) -> bool:
        """Whether this is a write request.

        Returns:
            bool: True if this is a write.
        """
        return self.fields["operation"].value == self.WRITE

    @property
    def is_safeload(self) -> bool:
        """Whether this is a safeload write request.

        Returns:
            bool: True if this is a safeload request.
        """
        return self.fields["operation"].value == self.WRITE and self.fields["safeload"].value == 1

    @property
    def is_read_request(self) -> bool:
        """Whether this is a read request.

        Returns:
            bool: True if this is a read request.
        """
        return self.fields["operation"].value == self.READ_REQUEST

    @property
    def is_read_response(self) -> bool:
        """Whether this is a read response.

        Returns:
            bool: True if this is a read response.
        """
        return self.fields["operation"].value == self.READ_RESPONSE

    def as_bytes(self) -> bytes:
        """Return the bytes representation of the current header.

        Returns:
            bytes: the assembled header
        """
        return self.fields.as_bytes()

    def parse(self, data: bytes):
        """Parse a header and populate field values.

        Args:
            data (bytes): The data to parse.
        """
        if len(data) != self.fields.size:
            raise ValueError(f"Input data needs to be exactly {self.fields.size} bytes long!")

        for field in self.fields:
            field.value = bytes_to_int(data, field.offset, field.size)

    def __setitem__(self, name: str, value: int):
        """Set a field value.

        Args:
            name (str): Field name.
            value (int): Field value.
        """
        valid_names = [field.name for field in self.fields]

        if name not in valid_names:
            raise ValueError(f"Invalid field name {name}; valid names are {', '.join(valid_names)}")

        self.fields[name].value = value

    def __getitem__(self, name: str) -> int:
        """Get a field value.

        Args:
            name (str): Field name.

        Returns:
            int: The field value.
        """
        return self.fields[name].value


_REGISTRY: Dict[int, Dict[str, Type[SigmaProtocolHeader]]] = {}


def _register(operation: int, chip_type: str, header_class: Type[SigmaProtocolHeader]):
    """Add a class to the header registry for the given operation and chip type.

    Args:
        operation (int): The operation code.
        chip_type (str): Chip type.
        header_class (Type[SigmaProtocolHeader]): Header class.
    """
    if operation not in _REGISTRY:
        _REGISTRY[operation] = {}

    _REGISTRY[operation][chip_type] = header_class


def _get_header(dsp_type: str, operation: int) -> SigmaProtocolHeader:
    """Select the appropriate header class for the chip type and operation requested.

    Args:
        dsp_type (str): One of the supported DSP types.
        operation (int): The operation code.

    Returns:
        SigmaProtocolHeader: The header object.
    """
    if operation not in _REGISTRY:
        valid_opcodes = ", ".join([f"0x{op:02x}" for op in _REGISTRY])
        raise ValueError(f"Invalid operation 0x{operation:02x}; valid operations are {valid_opcodes}.")

    if dsp_type not in _REGISTRY[operation]:
        raise ValueError(f"DSP type {dsp_type} is not supported.")

    header = _REGISTRY[operation][dsp_type]()
    header["operation"] = operation
    return header


class Adau1701WriteHeader(SigmaProtocolHeader):
    """Write header for the TCPIP1701 protocol.

    This is used for ADAU1701 and ADAU1401.
    """

    fields = Fields(
        [
            Field("operation", 0, 1),
            Field("safeload", 1, 1),
            Field("channel", 2, 1),
            Field("total_length", 3, 2),
            Field("chip_address", 5, 1),
            Field("data_length", 6, 2),
            Field("address", 8, 2),
        ]
    )


class Adau1701ReadRequestHeader(SigmaProtocolHeader):
    """Read request header for the TCPIP1701 protocol.

    This is used for ADAU1701 and ADAU1401 chips.
    """

    fields = Fields(
        [
            Field("operation", 0, 1),
            Field("total_length", 1, 2),
            Field("chip_address", 3, 1),
            Field("data_length", 4, 2),
            Field("address", 6, 2),
        ]
    )


class Adau1701ReadResponseHeader(SigmaProtocolHeader):
    """Read response header for the TCPIP1701 protocol.

    This is used for ADAU1701 and ADAU1401.
    """

    fields = Fields(
        [
            Field("operation", 0, 1),
            Field("total_length", 1, 2),
            Field("address", 3, 1),
        ]
    )


_register(SigmaProtocolHeader.WRITE, "adau1701", Adau1701WriteHeader)
_register(SigmaProtocolHeader.READ_REQUEST, "adau1701", Adau1701ReadRequestHeader)
_register(SigmaProtocolHeader.READ_RESPONSE, "adau1701", Adau1701ReadResponseHeader)


class Adau145xWriteHeader(SigmaProtocolHeader):
    """Write header for the TCPIPADAU144x/5x/6x protocols.

    This is presumably used for ADAU144x, ADAU145x and ADAU146x chips.
    """

    fields = Fields(
        [
            Field("operation", 0, 1),
            Field("safeload", 1, 1),
            Field("channel", 2, 1),
            Field("total_length", 3, 4),
            Field("chip_address", 7, 1),
            Field("data_length", 8, 4),
            Field("address", 12, 2),
        ]
    )


class Adau145xReadRequestHeader(SigmaProtocolHeader):
    """Read request header for the TCPIPADAU144x/5x/6x protocols.

    This is presumably used for ADAU144x, ADAU145x and ADAU146x chips.
    """

    fields = Fields(
        [
            Field("operation", 0, 1),
            Field("total_length", 1, 4),
            Field("chip_address", 5, 1),
            Field("data_length", 6, 4),
            Field("address", 10, 2),
            Field("reserved", 12, 2),
        ]
    )


class Adau145xReadResponseHeader(SigmaProtocolHeader):
    """Read response header for the TCPIPADAU144x/5x/6x protocols.

    This is presumably used for ADAU144x, ADAU145x and ADAU146x chips.
    """

    fields = Fields(
        [
            Field("operation", 0, 1),
            Field("total_length", 1, 4),
            Field("chip_address", 5, 1),
            Field("data_length", 6, 4),
            Field("address", 10, 2),
            Field("success", 12, 1),
            Field("reserved", 13, 1),
        ]
    )


_register(SigmaProtocolHeader.WRITE, "adau144x", Adau145xWriteHeader)
_register(SigmaProtocolHeader.READ_REQUEST, "adau144x", Adau145xReadRequestHeader)
_register(SigmaProtocolHeader.READ_RESPONSE, "adau144x", Adau145xReadResponseHeader)
_register(SigmaProtocolHeader.WRITE, "adau145x", Adau145xWriteHeader)
_register(SigmaProtocolHeader.READ_REQUEST, "adau145x", Adau145xReadRequestHeader)
_register(SigmaProtocolHeader.READ_RESPONSE, "adau145x", Adau145xReadResponseHeader)
_register(SigmaProtocolHeader.WRITE, "adau146x", Adau145xWriteHeader)
_register(SigmaProtocolHeader.READ_REQUEST, "adau146x", Adau145xReadRequestHeader)
_register(SigmaProtocolHeader.READ_RESPONSE, "adau146x", Adau145xReadResponseHeader)


class SigmaProtocolPacket:
    """Communication packet for talking to SigmaStudio.

    This class handles all of the payload decoding / encoding.
    """

    dsp_type: str
    header: SigmaProtocolHeader
    payload: bytearray

    def __init__(self, dsp_type: str):
        """Initialize the packet.

        Args:
            dsp_type (str): DSP type.
        """
        self.dsp_type = dsp_type

    def init_from_payload(self, operation: int, payload: bytearray, header_defaults: Union[Fields, None] = None):
        """Initialize the packet with a new header and the provided payload.

        Args:
            operation (int): Operation code; most often SigmaStudioPacket.READ_RESPONSE
            payload (bytearray): Payload.
            header_defaults (Union[Fields, None]): Default values for header fields; useful when setting chip address,
                etc. from therequest headers
        """
        self.header = _get_header(self.dsp_type, operation)
        self.payload = payload
        self.header["total_length"] = self.header.fields.size + len(payload)

        # adau145x and friends have some additional response fields
        if "data_length" in self.header.fields:
            self.header["data_length"] = len(payload)

        if header_defaults is None:
            return

        for field in self.header.fields:
            if field.name in ["operation", "total_length", "data_length", "success"]:
                continue

            if field.name in header_defaults:
                self.header.fields[field.name].value = header_defaults[field.name].value

    def init_from_network(self, request_handler: "SigmaStudioRequestHandler"):
        """Fetch data from the request handler.

        Args:
            request_handler (SigmaStudioRequestHandler): The request handler that deals with the network.
        """
        # first get the operation code
        header_bytes: bytearray = request_handler.read(1)

        # get the appropriate header object
        self.header = _get_header(self.dsp_type, header_bytes[0])

        # finally, read the rest of the header and parse it
        header_bytes.extend(request_handler.read(self.header.fields.size - 1))
        self.header.parse(header_bytes)

        if self.header.is_write_request:
            # we have a payload
            self.payload = request_handler.read(self.header["data_length"])

    @property
    def as_bytes(self) -> bytes:
        """Return the whole packet as bytes, ready for sending over the network.

        This also handles all of the math involved in setting up the header.

        Returns:
            bytes: header and payload combined
        """
        buffer = bytearray(self.header.as_bytes())

        buffer.extend(self.payload)

        return bytes(buffer)



"""This module contains the base classes for SigmaStudio protocol handling."""
import socketserver
from dataclasses import dataclass
from multiprocessing import Pipe


@dataclass(frozen=True)
class WriteRequest:
    """SigmaStudio requests to write data to the DSP."""

    address: int
    data: bytes


class SafeloadRequest(WriteRequest):
    """SigmaStudio requests to write data to the DSP using safeload."""


@dataclass(frozen=True)
class ReadRequest:
    """SigmaStudio requests to read data from the DSP."""

    address: int
    length: int


@dataclass(frozen=True)
class ReadResponse:
    """SigmaStudio is sent a response to its ReadRequest."""

    data: bytes


class ThreadedTCPServer(socketserver.ThreadingTCPServer):
    """The threaded TCP server that is used for communicating with SigmaStudio.

    It will be instantiated by SigmaStudioInterface. General server settings can be adjusted below.
    """

    allow_reuse_address = True

    def __init__(self, *args, **kwargs):
        """Initialize the ThreadedTCPServer with a Pipe for communicating with the TCP server worker thread."""
        self.pipe_end_owner, self.pipe_end_user = Pipe()

        super().__init__(*args, **kwargs)
